@@ -25,6 +25,8 @@
 #include <signal.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <dirent.h>
+#include <errno.h>
 
 const int RAISE = 1;
 const int LOWER = 0;
@@ -40,8 +42,10 @@ int local_priority = -1;
 static struct debconfclient *debconf;
 
 static int di_config_package(di_system_package *p,
-			     int (*virtfunc)(di_system_package *),
-			     int (*walkfunc)(di_system_package *));
+			     int (*virtfunc)(di_system_package *));
+
+static void modify_debconf_priority (int raise_or_lower);
+
 /*
  * qsort comparison function (sort by menu item values, fall back to
  * lexical sort to resolve ties deterministically).
@@ -148,43 +152,12 @@ get_default_menu_item(di_slist *list)
 	return NULL;
 }
 
-/* Return the text of the menu entry for PACKAGE, translated to
-   LANGUAGE if possible.  */
-static size_t menu_entry(struct debconfclient *debconf, char *language, di_system_package *package, char *buf, size_t size)
+/* Return the text of the menu entry for PACKAGE.  */
+static size_t menu_entry(struct debconfclient *debconf, di_system_package *package, char *buf, size_t size)
 {
 	char question[256];
 
 	snprintf(question, sizeof(question), "debian-installer/%s/title", package->p.package);
-	if (language) {
-		/*  256 + strlen("Description-.UTF-8")  */
-		char field[256+18];
-		char *lang, *end, *und;
-
-		strcpy(field, "Description-");
-		lang = language;
-		do {
-			end = strchr(lang, ':');
-			if (end == NULL)
-				end = lang + strlen(lang);
-		        strncpy(field+12, lang, end-lang);
-		        strcpy(field+12+(end-lang), ".UTF-8");
-			if (!debconf_metaget(debconf, question, field)) {
-				strncpy(buf, debconf->value, size);
-				return strlen (buf);
-			}
-			// Was the language of the form xx_YY ? 
-			und = strchr(lang, '_');
-			if (und != NULL && (und - lang) < (end - lang)) {
-		        	strncpy(field+12, lang, und-lang);
-		        	strcpy(field+12+(und-lang), ".UTF-8");
-				if (!debconf_metaget(debconf, question, field)) {
-	                        	strncpy(buf, debconf->value, size);
-	                        	return strlen (buf);
-	                	}
-			}
-			lang = end+1;
-		} while (*end != '\0');
-	}
 	if (!debconf_metaget(debconf, question, "Description")) {
 		strncpy(buf, debconf->value, size);
 		return strlen (buf);
@@ -200,7 +173,6 @@ static size_t menu_entry(struct debconfclient *debconf, char *language, di_syste
 
 /* Displays the main menu via debconf and returns the selected menu item. */
 di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *allocator) {
-	char *language = NULL;
 	di_system_package **package_array, *p;
 	di_slist *list;
 	di_slist_node *node;
@@ -209,9 +181,6 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 	char buf[256], *menu, *s;
 	int menu_size, menu_used, size;
 
-	if (debconf_get(debconf,"debian-installer/language") == 0 && debconf->value)
-		language = strdup(debconf->value);
-
 	for (node = packages->list.head; node; node = node->next) {
 		p = node->data;
 		if (((di_system_package *)p)->installer_menu_item)
@@ -244,7 +213,7 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		if (!p->installer_menu_item ||
 		    !isinstallable(p))
 			continue;
-		size = menu_entry(debconf, language, p, buf, sizeof (buf));
+		size = menu_entry(debconf, p, buf, sizeof (buf));
 		if (menu_used + size + 2 > menu_size)
 		{
 			menu_size += 1024;
@@ -263,10 +232,14 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 	debconf_capb(debconf);
 	debconf_subst(debconf, MAIN_MENU, "MENU", menu);
 	if (menudefault) {
-		menu_entry(debconf, language, menudefault, buf, sizeof (buf));
+		menu_entry(debconf, menudefault, buf, sizeof (buf));
 		debconf_set(debconf, MAIN_MENU, buf);
 	}
-	debconf_input(debconf, "medium", MAIN_MENU);
+	else {
+		di_log(DI_LOG_LEVEL_INFO, "no default menu item"); 
+		modify_debconf_priority(LOWER);
+	}
+	debconf_input(debconf, MENU_PRIORITY, MAIN_MENU);
 	debconf_go(debconf);
 	debconf_get(debconf, MAIN_MENU);
 	s = strdup(debconf->value);
@@ -274,13 +247,13 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 	/* Figure out which menu item was selected. */
 	for (i = 0; i < num; i++) {
 		p = package_array[i];
-		menu_entry(debconf, language, p, buf, sizeof (buf));
+		menu_entry(debconf, p, buf, sizeof (buf));
 		if (strcmp(buf, s) == 0) {
 			ret = p;
 			break;
 		}
 	}
-	
+
 	if (! ret) {
 		/* This could happen because of a debconf protocol problem
 		 * (for example, leading whitespace in menu items can
@@ -288,15 +261,13 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		 * problem. */
 		di_log(DI_LOG_LEVEL_WARNING, "Internal error! Cannot find \"%s\" in menu.", s);
 	}
-	
-	free(language);
+
+	free(menu);
 	free(package_array);
 
 	return ret;
 }
 
-static int check_special(di_system_package *p);
-
 /*
  * Satisfy the dependencies of a virtual package. Its dependencies
  * that actually provide the package are presented in a debconf select
@@ -306,14 +277,10 @@ static int check_special(di_system_package *p);
 static int satisfy_virtual(di_system_package *p) {
 	di_slist_node *node;
 	di_system_package *dep, *defpkg = NULL;
-	char *language = NULL;
 	char buf[256], *menu, *s = NULL;
 	size_t menu_size, menu_used, size;
 	int is_menu_item = 0;
 
-	if (debconf_get(debconf, "debian-installer/language") == 0 && debconf->value)
-		language = strdup(debconf->value);
-
 	menu = di_malloc(1024);
 	menu[0] = '\0';
 	menu_size = 1024;
@@ -327,9 +294,14 @@ static int satisfy_virtual(di_system_package *p) {
 		if (d->type == di_package_dependency_type_depends) {
 			/* Non-providing dependency */
 			di_log(DI_LOG_LEVEL_DEBUG, "non-providing dependency from %s to %s", p->p.package, dep->p.package);
-			if (dep->p.status != di_package_status_installed &&
-			    di_config_package(dep, satisfy_virtual, check_special) == -1)
-				return -1;
+			if (dep->p.status != di_package_status_installed) {
+				switch (di_config_package(dep, satisfy_virtual)) {
+					case -1:
+						return -1;
+					case EXIT_BACKUP:
+						return EXIT_BACKUP;
+				}
+			}
 			continue;
 		}
 		if (d->type != di_package_dependency_type_reverse_provides)
@@ -349,7 +321,7 @@ static int satisfy_virtual(di_system_package *p) {
 		if (dep->installer_menu_item)
 			is_menu_item = 1;
 
-		size = menu_entry(debconf, language, dep, buf, sizeof (buf));
+		size = menu_entry(debconf, dep, buf, sizeof (buf));
 		if (menu_used + size + 2 > menu_size)
 		{
 			menu_size += 1024;
@@ -362,12 +334,16 @@ static int satisfy_virtual(di_system_package *p) {
 			temp = di_malloc (menu_size);
 			strcpy(temp, menu);
 			strcpy(menu, buf);
-			strcat(menu, ", ");
-			strcat(menu, temp);
+			if (strlen(temp)) {
+				strcat(menu, ", ");
+				strcat(menu, temp);
+			}
 			di_free(temp);
 		} else {
+			if (strlen(menu)) {
+				strcat(menu, ", ");
+			}
 			strcat(menu, buf);
-			strcat(menu, ", ");
 		}
 		menu_used += size + 2;
 	}
@@ -378,7 +354,7 @@ static int satisfy_virtual(di_system_package *p) {
 			char *priority = "medium";
 			/* Only let the user choose if one of them is a menu item */
 			if (defpkg != NULL) {
-				menu_entry(debconf, language, defpkg, buf, sizeof(buf));
+				menu_entry(debconf, defpkg, buf, sizeof(buf));
 				debconf_set(debconf, MISSING_PROVIDE, buf);
 			} else {
 				/* TODO: How to figure out a default? */
@@ -397,13 +373,17 @@ static int satisfy_virtual(di_system_package *p) {
 		for (node = p->p.depends.head; node; node = node->next) {
 			di_package_dependency *d = node->data;
 			dep = (di_system_package *)d->ptr;
-			menu_entry(debconf, language, dep, buf, sizeof(buf));
+			menu_entry(debconf, dep, buf, sizeof(buf));
 			if (!is_menu_item || strcmp(s, buf) == 0) {
 				/* Ick. If we have a menu item it has to match the
 				 * debconf choice, otherwise we configure all of
 				 * the providing packages */
-				if (di_config_package(dep, satisfy_virtual, check_special) == -1)
-					return -1;
+				switch (di_config_package(dep, satisfy_virtual)) {
+					case -1:
+						return -1;
+					case EXIT_BACKUP:
+						return EXIT_BACKUP;
+				}
 				if (is_menu_item)
 					break;
 			}
@@ -411,39 +391,12 @@ static int satisfy_virtual(di_system_package *p) {
 	}
 	free(menu);
 	free(s);
-	free(language);
 	/* It doesn't make sense to configure virtual packages,
 	 * since they are, well, virtual. */
 	p->p.status = di_package_status_installed;
 	return 1;
 }
 
-static void update_language (void) {
-	debconf_get(debconf, "debian-installer/language");
-	if (*debconf->value != 0)
-		setenv("LANGUAGE", debconf->value, 1);
-}
-
-static int
-check_special(di_system_package *p)
-{
-	di_slist_node *node;
-
-	/*
-	 * A language selection package must provide the virtual
-	 * package 'language-selected'.
-	 * The LANGUAGE environment variable must be updated
-	 */
-	for (node = p->p.depends.head; node; node = node->next) {
-		di_package_dependency *d = node->data;
-		if (d->type == di_package_dependency_type_provides && strcmp(d->ptr->package, "language-selected") == 0) {
-			update_language();
-			break;
-		}
-	}
-	return 0;
-}
-
 static void set_package_title(di_system_package *p) {
 	char *title;
 
@@ -456,9 +409,9 @@ static void set_package_title(di_system_package *p) {
 }
 
 static int do_menu_item(di_system_package *p) {
-	di_log(DI_LOG_LEVEL_DEBUG, "Menu item '%s' selected", p->p.package);
+	di_log(DI_LOG_LEVEL_INFO, "Menu item '%s' selected", p->p.package);
 
-	return di_config_package(p, satisfy_virtual, check_special);
+	return di_config_package(p, satisfy_virtual);
 }
 
 static char *debconf_priorities[] =
@@ -472,9 +425,20 @@ static char *debconf_priorities[] =
 #define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
 static void modify_debconf_priority (int raise_or_lower) {
 	int pri, i;
+	static int menu_pri = -1;
 	const char *template = "debconf/priority";
 	debconf_get(debconf, template);
 	
+	if (menu_pri == -1) {
+		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
+			if (0 == strcmp(MENU_PRIORITY,
+					debconf_priorities[i]) ) {
+				menu_pri = i;
+				break;
+			}
+		}
+	}
+
 	pri = 1;
 	if ( debconf->value ) {
 		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
@@ -485,8 +449,14 @@ static void modify_debconf_priority (int raise_or_lower) {
 			}
 		}
 	}
-	if (raise_or_lower == LOWER)
+	if (raise_or_lower == LOWER) {
 		--pri;
+		/* Make sure the menu is always displayed after a single
+		 * backing up.
+		 */
+		if (menu_pri != -1 && pri > menu_pri)
+			pri = menu_pri;
+	}
 	else if (raise_or_lower == RAISE)
 		++pri;
 	if (0 > pri)
@@ -494,9 +464,10 @@ static void modify_debconf_priority (int raise_or_lower) {
 	if (pri > default_priority)
 		pri = default_priority;
 	if (local_priority != pri) {
-		di_log(DI_LOG_LEVEL_INFO, "Modifying debconf priority limit from '%s' to '%s'",
-			debconf->value ? debconf->value : "(null)",
-			debconf_priorities[pri] ? debconf_priorities[pri] : "(null)");
+		if (local_priority > -1)
+			di_log(DI_LOG_LEVEL_INFO, "Modifying debconf priority limit from '%s' to '%s'",
+				debconf->value ? debconf->value : "(null)",
+				debconf_priorities[pri] ? debconf_priorities[pri] : "(null)");
 		local_priority = pri;
 	    
 		debconf_set(debconf, template, debconf_priorities[pri]);
@@ -519,29 +490,79 @@ static void adjust_default_priority (void) {
 		}
 	}
 	
-	if ( pri != local_priority ) {
-		di_log(DI_LOG_LEVEL_INFO, "Priority changed externally, setting main-menu default to '%s' (%s)",
-			debconf_priorities[pri] ? debconf_priorities[pri] : "(null)", debconf->value);
+	if ( pri != local_priority) {
+		if (local_priority > -1)
+			di_log(DI_LOG_LEVEL_INFO, "Priority changed externally, setting main-menu default to '%s' (%s)",
+				debconf_priorities[pri] ? debconf_priorities[pri] : "(null)", debconf->value);
 		local_priority = pri;
 		default_priority = pri;
 	}
 }
 
 void notify_user_of_failure (di_system_package *p) {
-	char *language = NULL;
 	char buf[256];
 	
 	debconf_capb(debconf);
-	debconf_get(debconf,"debian-installer/language");
-	if (debconf->value)
-		language = strdup(debconf->value);
-	menu_entry(debconf, language, p, buf, sizeof (buf));
+	menu_entry(debconf, p, buf, sizeof (buf));
 	debconf_subst(debconf, ITEM_FAILURE, "ITEM", buf);
 	debconf_input(debconf, "critical", ITEM_FAILURE);
 	debconf_go(debconf);
 	debconf_capb(debconf, "backup");
 }
 
+/* Cheap-and-cheerful run-parts-a-like for /lib/main-menu.d. Allows packages
+ * to register scripts to be run at main-menu startup that need to share
+ * main-menu's debconf frontend but that don't merit a menu item, such as
+ * setting an info message.
+ */
+static void menu_startup (void) {
+	struct dirent **namelist;
+	int entries, i;
+
+	/* scandir() isn't POSIX, but it makes things easy. */
+	entries = scandir(MAIN_MENU_DIR, &namelist, NULL, alphasort);
+	if (entries < 0)
+		return;
+
+	for (i = 0; i < entries; ++i) {
+		size_t len;
+		char *filename;
+		struct stat st;
+		int ret;
+
+		if (strcmp(namelist[i]->d_name, ".") == 0 || strcmp(namelist[i]->d_name, "..") == 0)
+			continue;
+
+		/* sizeof(MAIN_MENU_DIR) includes trailing \0 */
+		len = sizeof(MAIN_MENU_DIR) + 1 + strlen(namelist[i]->d_name);
+		filename = di_new(char, len);
+		snprintf(filename, len, "%s/%s", MAIN_MENU_DIR, namelist[i]->d_name);
+
+		if (stat(filename, &st) != 0) {
+			di_log(DI_LOG_LEVEL_WARNING, "Can't stat %s (%s)", filename, strerror(errno));
+			di_free(filename);
+			continue;
+		}
+		if (!S_ISREG(st.st_mode)) {
+			di_log(DI_LOG_LEVEL_WARNING, "%s is not a regular file", filename);
+			di_free(filename);
+			continue;
+		}
+		if (access(filename, X_OK) != 0) {
+			di_log(DI_LOG_LEVEL_WARNING, "%s is not executable", filename);
+			di_free(filename);
+			continue;
+		}
+
+		//di_log(DI_LOG_LEVEL_DEBUG, "Executing %s", filename);
+		ret = system(filename);
+		if (ret != 0)
+			di_log(DI_LOG_LEVEL_WARNING, "%s exited with status %d", filename, ret);
+
+		di_free(filename);
+	}
+}
+
 int main (int argc __attribute__ ((unused)), char **argv) {
 	di_system_package *p;
 	di_packages *packages;
@@ -554,6 +575,14 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	/* Tell udpkg to shut up. */
 	setenv("UDPKG_QUIET", "y", 1);
 
+	/* Make cdebconf honour currently set language */
+	const char *template = "debconf/language";
+	if (debconf_get(debconf, template) == CMD_SUCCESS && 
+	    debconf->value && *debconf->value)
+		debconf_set(debconf, template, debconf->value);
+
+	menu_startup();
+
 	allocator = di_system_packages_allocator_alloc ();
 	packages = di_system_packages_status_read_file(DI_SYSTEM_DPKG_STATUSFILE, allocator);
 	while ((p=show_main_menu(packages, allocator))) {
@@ -598,17 +627,16 @@ int main (int argc __attribute__ ((unused)), char **argv) {
  * This is done depth-first.
  */
 static int di_config_package(di_system_package *p,
-			     int (*virtfunc)(di_system_package *),
-			     int (*walkfunc)(di_system_package *)) {
+			     int (*virtfunc)(di_system_package *)) {
 	char *configcommand;
 	int ret;
 	di_slist_node *node;
 	di_system_package *dep;
 
-	di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d\n", p->p.package, p->p.status);
+	//di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d\n", p->p.package, p->p.status);
 
 	if (p->p.type == di_package_type_virtual_package) {
-		di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s\n", p->p.package);
+		//di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s\n", p->p.package);
 		if (virtfunc)
 			return virtfunc(p);
 		else
@@ -625,8 +653,12 @@ static int di_config_package(di_system_package *p,
 		if (d->type != di_package_dependency_type_depends)
 			continue;
 		/* Recursively configure this package */
-		if (di_config_package(dep, virtfunc, walkfunc) == -1)
-			return -1;
+		switch (di_config_package(dep, virtfunc)) {
+			case -1:
+				return -1;
+			case EXIT_BACKUP:
+				return EXIT_BACKUP;
+		}
 	}
 
 	set_package_title(p);
@@ -639,8 +671,6 @@ static int di_config_package(di_system_package *p,
 	switch (ret) {
 		case EXIT_OK:
 			p->p.status = di_package_status_installed;
-			if (walkfunc != NULL)
-				walkfunc(p);
 			break;
 		default:
 			di_log(DI_LOG_LEVEL_WARNING, "Configuring '%s' failed with error code %d", p->p.package, ret);
