@@ -12,6 +12,7 @@
 #define MAIN_MENU	"debian-installer/main-menu"
 #define MISSING_PROVIDE "debian-installer/missing-provide"
 #define ITEM_FAILURE	"debian-installer/main-menu/item-failure"
+#define MAIN_MENU_DIR	"/lib/main-menu.d"
 
 #include <debian-installer.h>
 
@@ -23,5 +24,8 @@ enum
 	EXIT_BACKUP		= 10,
 };
 
+/* Priority at which the menu is displayed */
+#define MENU_PRIORITY	"medium"
+
 /* vim: noexpandtab sw=8
  */
